/*-
 * Copyright (c) 2006 Verdens Gang AS
 * Copyright (c) 2006-2008 Linpro AS
 * All rights reserved.
 *
 * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cache_fetch.c 2564 2008-03-08 13:38:23Z des $
 */

#include "config.h"

#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "shmlog.h"
#include "cache.h"
#include "heritage.h"
#include "stevedore.h"

/*--------------------------------------------------------------------*/

static int
fetch_straight(struct sess *sp, struct http_conn *htc, const char *b)
{
	int i;
	unsigned char *p;
	uintmax_t cll;
	unsigned cl;
	struct storage *st;

	cll = strtoumax(b, NULL, 0);
	if (cll == 0)
		return (0);

	cl = (unsigned)cll;
	assert((uintmax_t)cl == cll); /* Protect against bogusly large values */

	st = STV_alloc(sp, cl);
	VTAILQ_INSERT_TAIL(&sp->obj->store, st, list);
	st->len = cl;
	sp->obj->len = cl;
	p = st->ptr;

	i = fcntl(htc->fd, F_GETFL);		/* XXX ? */
	i &= ~O_NONBLOCK;
	i = fcntl(htc->fd, F_SETFL, i);

	while (cl > 0) {
		i = HTC_Read(htc, p, cl);
		if (i <= 0)
			return (-1);
		p += i;
		cl -= i;
	}
	return (0);
}

/*--------------------------------------------------------------------*/
/* XXX: Cleanup.  It must be possible somehow :-( */

static int
fetch_chunked(struct sess *sp, struct http_conn *htc)
{
	int i;
	char *q;
	struct storage *st;
	unsigned u, v, w;
	char buf[20];		/* XXX: arbitrary */
	char *bp, *be;

	be = buf + sizeof buf - 1;
	bp = buf;
	st = NULL;
	u = 0;
	while (1) {
		/* Try to parse buf as a chunk length */
		*bp = '\0';
		u = strtoul(buf, &q, 16);

		/* Skip trailing whitespace */
		if (q != NULL && q > buf) {
			while (*q == '\t' || *q == ' ')
				q++;
			if (*q == '\r')
				q++;
		}

		/* If we didn't succeed, add to buffer, try again */
		if (q == NULL || q == buf || *q != '\n') {
			xxxassert(be > bp);
			/*
			 * The sematics we need here is "read until you have
			 * received at least one character, but feel free to
			 * return up to (be-bp) if they are available, but do
			 * not wait for those extra characters.
			 *
			 * The canonical way to do that is to do a blocking
			 * read(2) of one char, then change to nonblocking,
			 * read as many as we find, then change back to
			 * blocking reads again.
			 *
			 * Hardly much more efficient and certainly a good
			 * deal more complex than reading a single character
			 * at a time.
			 */
			i = HTC_Read(htc, bp, 1);
			if (i <= 0)
				return (-1);
			bp += i;
			continue;
		}

		/* Skip NL */
		q++;

		/* Last chunk is zero bytes long */
		if (u == 0)
			break;

		while (u > 0) {

			/* Get some storage if we don't have any */
			if (st == NULL || st->len == st->space) {
				v = u;
				if (u < params->fetch_chunksize * 1024)
					v = params->fetch_chunksize * 1024;
				st = STV_alloc(sp, v);
				VTAILQ_INSERT_TAIL(&sp->obj->store, st, list);
			}
			v = st->space - st->len;
			if (v > u)
				v = u;

			/* Handle anything left in our buffer first */
			w = pdiff(q, bp);
			if (w > v)
				w = v;
			if (w != 0) {
				memcpy(st->ptr + st->len, q, w);
				st->len += w;
				sp->obj->len += w;
				u -= w;
				v -= w;
				q += w;
			}
			if (u == 0)
				break;
			if (v == 0)
				continue;

			/* Pick up the rest of this chunk */
			while (v > 0) {
				i = HTC_Read(htc, st->ptr + st->len, v);
				if (i <= 0)
					return (-1);
				st->len += i;
				sp->obj->len += i;
				u -= i;
				v -= i;
			}
		}
		assert(u == 0);

		/* We might still have stuff in our buffer */
		v = pdiff(q, bp);
		if (v > 0)
			memcpy(buf, q, v);
		q = bp = buf + v;
	}

	if (st != NULL && st->len == 0) {
		VTAILQ_REMOVE(&sp->obj->store, st, list);
		STV_free(st);
	} else if (st != NULL)
		STV_trim(st, st->len);
	return (0);
}


/*--------------------------------------------------------------------*/

static int
fetch_eof(struct sess *sp, struct http_conn *htc)
{
	int i;
	unsigned char *p;
	struct storage *st;
	unsigned v;

	i = fcntl(htc->fd, F_GETFL);		/* XXX ? */
	i &= ~O_NONBLOCK;
	i = fcntl(htc->fd, F_SETFL, i);

	p = NULL;
	v = 0;
	st = NULL;
	while (1) {
		if (v == 0) {
			st = STV_alloc(sp, params->fetch_chunksize * 1024);
			VTAILQ_INSERT_TAIL(&sp->obj->store, st, list);
			p = st->ptr + st->len;
			v = st->space - st->len;
		}
		AN(p);
		AN(st);
		i = HTC_Read(htc, p, v);
		if (i < 0)
			return (-1);
		if (i == 0)
			break;
		p += i;
		v -= i;
		st->len += i;
		sp->obj->len += i;
	}

	if (st->len == 0) {
		VTAILQ_REMOVE(&sp->obj->store, st, list);
		STV_free(st);
	} else
		STV_trim(st, st->len);

	return (1);
}

/*--------------------------------------------------------------------*/

int
Fetch(struct sess *sp)
{
	struct vbe_conn *vc;
	struct worker *w;
	char *b;
	int cls;
	struct http *hp, *hp2;
	struct storage *st;
	struct bereq *bereq;
	int mklen, is_head;
	struct http_conn htc[1];
	int i;
	char *ptr, *endp;

	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
	CHECK_OBJ_NOTNULL(sp->wrk, WORKER_MAGIC);
	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);
	CHECK_OBJ_NOTNULL(sp->bereq, BEREQ_MAGIC);
	CHECK_OBJ_NOTNULL(sp->backend, BACKEND_MAGIC);
	assert(sp->obj->busy != 0);
	w = sp->wrk;
	bereq = sp->bereq;
	hp = bereq->http;
	is_head = (strcasecmp(http_GetReq(hp), "head") == 0);

	sp->obj->xid = sp->xid;

	/* Set up obj's workspace */
	st = sp->obj->objstore;
	WS_Init(sp->obj->ws_o, "obj", st->ptr + st->len, st->space - st->len);
	st->len = st->space;
	WS_Assert(sp->obj->ws_o);
	http_Setup(sp->obj->http, sp->obj->ws_o);

	vc = VBE_GetFd(sp);
	if (vc == NULL)
		return (__LINE__);
	WRK_Reset(w, &vc->fd);
	http_Write(w, hp, 0);

	/*
	 * If a POST request was passed to fetch, we must send any
	 * pipelined bytes to the backend as well
	 */
	if (http_GetHdr(sp->http, H_Content_Length, &ptr)) {
		unsigned long content_length;
		char buf[8192];
		int read;

		content_length = strtoul(ptr, &endp, 10);
		/* XXX should check result of conversion */
		while (content_length) {
			if (content_length > sizeof buf)
				read = sizeof buf;
			else
				read = content_length;
			read = HTC_Read(sp->htc, buf, read);
			WRK_Write(w, buf, read);
			if (WRK_Flush(w)) {
				VBE_UpdateHealth(sp, vc, -1);
				VBE_ClosedFd(sp->wrk, vc);
				return (__LINE__);
			}
			content_length -= read;
		}
	}

	if (WRK_Flush(w)) {
		VBE_UpdateHealth(sp, vc, -1);
		VBE_ClosedFd(sp->wrk, vc);
		/* XXX: other cleanup ? */
		return (__LINE__);
	}

	/* XXX is this the right place? */
	VSL_stats->backend_req++;

	HTC_Init(htc, bereq->ws, vc->fd);
	do
		i = HTC_Rx(htc);
	while (i == 0);

	if (i < 0) {
		VBE_UpdateHealth(sp, vc, -1);
		VBE_ClosedFd(sp->wrk, vc);
		/* XXX: other cleanup ? */
		return (__LINE__);
	}

	if (http_DissectResponse(sp->wrk, htc, hp)) {
		VBE_UpdateHealth(sp, vc, -2);
		VBE_ClosedFd(sp->wrk, vc);
		/* XXX: other cleanup ? */
		return (__LINE__);
	}

	sp->obj->entered = TIM_real();

	if (http_GetHdr(hp, H_Last_Modified, &b))
		sp->obj->last_modified = TIM_parse(b);

	/* Filter into object */
	hp2 = sp->obj->http;

	hp2->logtag = HTTP_Obj;
	http_CopyResp(hp2, hp);
	http_FilterFields(sp->wrk, sp->fd, hp2, hp, HTTPH_A_INS);
	http_CopyHome(sp->wrk, sp->fd, hp2);

	/* Determine if we have a body or not */
	cls = 0;
	mklen = 0;
	if (is_head) {
		/* nothing */
	} else if (http_GetHdr(hp, H_Content_Length, &b)) {
		cls = fetch_straight(sp, htc, b);
		mklen = 1;
	} else if (http_HdrIs(hp, H_Transfer_Encoding, "chunked")) {
		cls = fetch_chunked(sp, htc);
		mklen = 1;
	} else if (http_GetHdr(hp, H_Transfer_Encoding, &b)) {
		/* XXX: AUGH! */
		WSL(sp->wrk, SLT_Debug, vc->fd, "Invalid Transfer-Encoding");
		VBE_UpdateHealth(sp, vc, -3);
		VBE_ClosedFd(sp->wrk, vc);
		return (__LINE__);
	} else {
		switch (http_GetStatus(hp)) {
			case 200:
				cls = fetch_eof(sp, htc);
				mklen = 1;
				break;
			default:
				break;
		}
	}

	if (cls < 0) {
		/* XXX: Wouldn't this store automatically be released ? */
		while (!VTAILQ_EMPTY(&sp->obj->store)) {
			st = VTAILQ_FIRST(&sp->obj->store);
			VTAILQ_REMOVE(&sp->obj->store, st, list);
			STV_free(st);
		}
		VBE_UpdateHealth(sp, vc, -4);
		VBE_ClosedFd(sp->wrk, vc);
		return (__LINE__);
	}

	{
	/* Sanity check fetch methods accounting */
		unsigned uu;

		uu = 0;
		VTAILQ_FOREACH(st, &sp->obj->store, list)
			uu += st->len;
		assert(uu == sp->obj->len);
	}

	if (mklen > 0)
		http_PrintfHeader(sp->wrk, sp->fd, hp2,
		    "Content-Length: %u", sp->obj->len);

	if (http_GetHdr(hp, H_Connection, &b) && !strcasecmp(b, "close"))
		cls = 1;

	VBE_UpdateHealth(sp, vc, http_GetStatus(sp->bereq->http));

	if (cls)
		VBE_ClosedFd(sp->wrk, vc);
	else
		VBE_RecycleFd(sp->wrk, vc);

	return (0);
}
